@@ -37,6 +37,13 @@ static const int pin_to_gpio_rev1[27] = {-1, -1, -1, 0, -1, 1, -1, 4, 14, -1, 15
 static const int pin_to_gpio_rev2[27] = {-1, -1, -1, 2, -1, 3, -1, 4, 14, -1, 15, 17, 18, 27, -1, 22, 23, -1, 24, 10, -1, 9, 25, 11, 8, -1, 7};
 static const int (*pin_to_gpio)[27];
 
+// Board header info is shifted left 8 bits (leaves space for up to 255 channel ids per header)
+#define HEADER_P1 0<<8
+#define HEADER_P5 5<<8
+static const int gpio_to_pin_rev1[32] = {3, 5, -1, -1, 7, -1, -1, 26, 24, 21, 19, 23, -1, -1, 8, 10, -1, 11, 12, -1, -1, 13, 15, 16, 18, 22, -1, -1, -1, -1, -1, -1};
+static const int gpio_to_pin_rev2[32] = {-1, -1, 3, 5, 7, -1, -1, 26, 24, 21, 19, 23, -1, -1, 8, 10, -1, 11, 12, -1, -1, -1, 15, 16, 18, 22, -1, 15, 3 | HEADER_P5, 4 | HEADER_P5, 5 | HEADER_P5, 6 | HEADER_P5};
+static const int (*gpio_to_pin)[32];
+
 // Flag whether to show warnings
 static int gpio_warnings = 1;
 
@@ -54,12 +61,32 @@ static int gpio_direction[54];
 #define BCM          11
 static int gpio_mode = MODE_UNKNOWN;
 
+// Read /proc/cpuinfo once and keep the info at hand for further requests
 static void
 cache_rpi_revision(void)
 {
     revision_int = get_cpuinfo_revision(revision_hex);
 }
 
+// bcm_to_board() returns the pin for the supplied bcm_gpio_id or -1
+// if not a valid gpio-id. P5 pins are returned with | HEADER_P5, so
+// you can know the header with (retval >> 8) (either 0 or 5) and the
+// exact pin number with (retval & 255).
+static int
+bcm_to_board(int bcm_gpio_id)
+{
+    return *(*gpio_to_pin+bcm_gpio_id);
+}
+
+// channel_to_bcm() returns the bcm gpio id for the supplied channel
+// depending on current setmode. Only P1 header channels are supported.
+// To use P5 you need to use BCM gpio ids (`setmode(BCM)`).
+static int
+board_to_bcm(int board_pin_id)
+{
+    return *(*pin_to_gpio+board_pin_id);
+}
+
 // module_setup is run on import of the GPIO module and calls the setup() method in c_gpio.c
 static int
 module_setup(void)
@@ -106,35 +133,55 @@ py_cleanup(PyObject *self, PyObject *args)
     return Py_None;
 }
 
+
+// channel_to_gpio tries to convert the supplied channel-id to
+// a BCM GPIO ID based on current setmode. On error it sets the
+// Python error string and returns a value < 0.
 static int
-verify_input(int channel, int *gpio)
+channel_to_gpio(int channel)
 {
-    if (gpio_mode != BOARD && gpio_mode != BCM) {
+    int gpio;
+
+     if (gpio_mode != BOARD && gpio_mode != BCM) {
         PyErr_SetString(ModeNotSetException, "Please set pin numbering mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)");
-        return 0;
+        return -1;
     }
 
-    if ( (gpio_mode == BCM   && (channel < 0 || channel > 53)) ||
-         (gpio_mode == BOARD && (channel < 1 || channel > 26)) ) {
-        PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-        return 0;
+   if ( (gpio_mode == BCM && (channel < 0 || channel > 31)) ||
+        (gpio_mode == BOARD && (channel < 1 || channel > 26)) ) {
+        PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi (outside of range)");
+        return -2;
     }
 
     if (gpio_mode == BOARD) {
-        *gpio = *(*pin_to_gpio+channel);
-        if (*gpio == -1) {
-            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-            return 0;
+        if ((gpio = board_to_bcm(channel)) == -1) {
+            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi (not a valid pin)");
+            return -3;
         }
     } else {
         // gpio_mode == BCM
-        *gpio = channel;
+        gpio = channel;
+        if (bcm_to_board(gpio) == -1) {
+            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi (not a valid gpio)");
+            return -3;
+        }
     }
 
+    //printf("channel2bcm: %d -> %d", channel, gpio);
+    return gpio;
+}
+
+static int
+verify_input(int channel, int *gpio)
+{
+    if ((*gpio = channel_to_gpio(channel)) == -1)
+        return 0;
+
     if ((gpio_direction[*gpio] != INPUT) && (gpio_direction[*gpio] != OUTPUT)) {
         PyErr_SetString(WrongDirectionException, "GPIO channel has not been set up");
         return 0;
     }
+
     return 1;
 }
 
@@ -164,31 +211,8 @@ py_setup_channel(PyObject *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    if (gpio_mode != BOARD && gpio_mode != BCM) {
-        PyErr_SetString(ModeNotSetException, "Please set mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)");
+    if ((gpio = channel_to_gpio(channel)) < 0)
         return NULL;
-    }
-
-    if ( (gpio_mode == BCM && (channel < 0 || channel > 53))
-      || (gpio_mode == BOARD && (channel < 1 || channel > 26)) )
-    {
-        PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-        return NULL;
-    }
-
-    if (gpio_mode == BOARD)
-    {
-        gpio = *(*pin_to_gpio+channel);
-        if (gpio == -1)
-        {
-            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-            return NULL;
-        }
-    }
-    else // gpio_mode == BCM
-    {
-        gpio = channel;
-    }
 
     func = gpio_function(gpio);
     if (gpio_warnings &&                                      // warnings enabled and
@@ -199,8 +223,7 @@ py_setup_channel(PyObject *self, PyObject *args, PyObject *kwargs)
     }
 
 //    printf("Setup GPIO %d direction %d pud %d\n", gpio, direction, pud);
-    if (direction == OUTPUT && (initial == LOW || initial == HIGH))
-    {
+    if (direction == OUTPUT && (initial == LOW || initial == HIGH)) {
 //        printf("Writing intial value %d\n",initial);
         output_gpio(gpio, initial);
     }
@@ -220,35 +243,10 @@ py_output_gpio(PyObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "ii", &channel, &value))
         return NULL;
 
-    if (gpio_mode != BOARD && gpio_mode != BCM)
-    {
-        PyErr_SetString(ModeNotSetException, "Please set mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)");
+    if ((gpio = channel_to_gpio(channel)) < 0)
         return NULL;
-    }
 
-    if ( (gpio_mode == BCM && (channel < 0 || channel > 53))
-      || (gpio_mode == BOARD && (channel < 1 || channel > 26)) )
-    {
-        PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-        return NULL;
-    }
-
-    if (gpio_mode == BOARD)
-    {
-        gpio = *(*pin_to_gpio+channel);
-        if (gpio == -1)
-        {
-            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-            return NULL;
-        }
-    }
-    else // gpio_mode == BCM
-    {
-        gpio = channel;
-    }
-
-    if (gpio_direction[gpio] != OUTPUT)
-    {
+    if (gpio_direction[gpio] != OUTPUT) {
         PyErr_SetString(WrongDirectionException, "The GPIO channel has not been set up as an OUTPUT");
         return NULL;
     }
@@ -270,40 +268,35 @@ py_forceoutput_gpio(PyObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "ii", &channel, &value))
         return NULL;
 
-    if (gpio_mode != BOARD && gpio_mode != BCM)
-    {
-        PyErr_SetString(ModeNotSetException, "Please set mode using GPIO.setmode(GPIO.BOARD) or GPIO.setmode(GPIO.BCM)");
-        return NULL;
-    }
-
-    if ( (gpio_mode == BCM && (channel < 0 || channel > 53))
-      || (gpio_mode == BOARD && (channel < 1 || channel > 26)) )
-    {
-        PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-        return NULL;
-    }
-
-    if (gpio_mode == BOARD)
-    {
-        gpio = *(*pin_to_gpio+channel);
-        if (gpio == -1)
-        {
-            PyErr_SetString(InvalidChannelException, "The channel sent is invalid on a Raspberry Pi");
-            return NULL;
-        }
-    }
-    else // gpio_mode == BCM
-    {
-        gpio = channel;
-    }
-
 //    printf("Output GPIO %d value %d\n", gpio, value);
+    if ((gpio = channel_to_gpio(channel)) < 0)
+        return NULL;
+
     output_gpio(gpio, value);
 
     Py_INCREF(Py_None);
     return Py_None;
 }
 
+// python function output(channel, value) without direction check
+static PyObject*
+py_set_pullupdn(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    int gpio, channel;
+    int pud = PUD_OFF;
+
+    static char *kwlist[] = {"channel", "pull_up_down", NULL};
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "ii|ii", kwlist, &channel, &pud))
+        return NULL;
+
+    if ((gpio = channel_to_gpio(channel)) < 0)
+        return NULL;
+
+    // printf("Setting gpio %d PULLUPDN to %d", gpio, pud);
+    set_pullupdn(gpio, pud);
+    return Py_None;
+}
+
 // python function value = input(channel)
 static PyObject*
 py_input_gpio(PyObject *self, PyObject *args)
@@ -327,9 +320,12 @@ py_input_gpio(PyObject *self, PyObject *args)
 static PyObject*
 py_forceinput_gpio(PyObject *self, PyObject *args)
 {
-    int gpio;
+    int gpio, channel;
 
-    if (!PyArg_ParseTuple(args, "i", &gpio))
+    if (!PyArg_ParseTuple(args, "i", &channel))
+        return NULL;
+
+    if ((gpio = channel_to_gpio(channel)) < 0)
         return NULL;
 
     //printf("Input GPIO %d\n", gpio);
@@ -360,10 +356,13 @@ setmode(PyObject *self, PyObject *args)
 static PyObject*
 py_gpio_function(PyObject *self, PyObject *args)
 {
-    int gpio, f;
+    int gpio, channel, f;
     PyObject *func;
 
-    if (!PyArg_ParseTuple(args, "i", &gpio))
+    if (!PyArg_ParseTuple(args, "i", &channel))
+        return NULL;
+
+    if ((gpio = channel_to_gpio(channel)) < 0)
         return NULL;
 
     f = gpio_function(gpio);
@@ -386,18 +385,37 @@ py_setwarnings(PyObject *self, PyObject *args)
     return Py_None;
 }
 
+// channel2gpio binding
+static PyObject*
+py_channel_to_gpio(PyObject *self, PyObject *args)
+{
+    int channel, gpio;
+    PyObject *func;
+
+    if (!PyArg_ParseTuple(args, "i", &channel))
+        return NULL;
+
+    if ((gpio = channel_to_gpio(channel)) < 0)
+        return NULL;
+
+    func = Py_BuildValue("i", gpio);
+    return func;
+}
+
 PyMethodDef rpi_gpio_methods[] = {
     {"setup", (PyCFunction)py_setup_channel, METH_VARARGS | METH_KEYWORDS, "Set up the GPIO channel, direction and (optional) pull/up down control\nchannel    - Either: RPi board pin number (not BCM GPIO 00..nn number).  Pins start from 1\n                or     : BCM GPIO number\ndirection - INPUT or OUTPUT\n[pull_up_down] - PUD_OFF (default), PUD_UP or PUD_DOWN\n[initial]        - Initial value for an output channel"},
     {"cleanup", py_cleanup, METH_VARARGS, "Clean up by resetting all GPIO channels that have been used by this program\nto INPUT with no pullup/pulldown and no event detection"},
     {"output", py_output_gpio, METH_VARARGS, "Output to a GPIO channel"},
     {"input", py_input_gpio, METH_VARARGS, "Input from a GPIO channel"},
     {"setmode", setmode, METH_VARARGS, "Set up numbering mode to use for channels.\nBOARD - Use Raspberry Pi board numbers\nBCM    - Use Broadcom GPIO 00..nn numbers"},
-    {"gpio_function", py_gpio_function, METH_VARARGS, "Return the current GPIO function (IN, OUT, ALT0)"},
     {"setwarnings", py_setwarnings, METH_VARARGS, "Enable or disable warning messages"},
 
     // New methods in RPIO
     {"forceoutput", py_forceoutput_gpio, METH_VARARGS, "Force output to a GPIO channel, ignoring whether it has been set up before."},
     {"forceinput", py_forceinput_gpio, METH_VARARGS, "Force read input from a GPIO channel, ignoring whether it was set up before."},
+    {"set_pullupdn", (PyCFunction)py_set_pullupdn, METH_VARARGS | METH_KEYWORDS, "Set pullup or -down resistor on a GPIO channel."},
+    {"gpio_function", py_gpio_function, METH_VARARGS, "Return the current GPIO function (IN, OUT, ALT0)"},
+    {"channel_to_gpio", py_channel_to_gpio, METH_VARARGS, "Return BCM or BOARD id of channel (depending on current setmode)"},
     {NULL, NULL, 0, NULL}
 };
 
@@ -488,9 +506,11 @@ PyMODINIT_FUNC initGPIO(void)
 #endif
     } else if (revision_int == 1) {
         pin_to_gpio = &pin_to_gpio_rev1;
+        gpio_to_pin = &gpio_to_pin_rev1;
     } else {
         // assume revision 2
         pin_to_gpio = &pin_to_gpio_rev2;
+        gpio_to_pin = &gpio_to_pin_rev2;
     }
 
     rpi_revision = Py_BuildValue("i", revision_int);
@@ -499,7 +519,7 @@ PyMODINIT_FUNC initGPIO(void)
     rpi_revision_hex = Py_BuildValue("s", revision_hex);
     PyModule_AddObject(module, "RPI_REVISION_HEX", rpi_revision_hex);
 
-    version = Py_BuildValue("s", "0.7.1/0.4.2a");
+    version = Py_BuildValue("s", "0.8.0/0.4.2a");
     PyModule_AddObject(module, "VERSION_GPIO", version);
 
     // set up mmaped areas
